/*
 *  Copyright  Northwoods Software Corporation, 2000-2007. All Rights
 *  Reserved.
 *
 *  Restricted Rights: Use, duplication, or disclosure by the U.S.
 *  Government is subject to restrictions as set forth in subparagraph
 *  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
 *  52.227-14 Alt. III, as applicable.
 *
 */

package com.nwoods.jgo.examples;

import org.eclipse.swt.*;
import org.eclipse.swt.events.*;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.layout.*;
import org.eclipse.swt.widgets.*;
import com.nwoods.jgo.*;

/**
 * This is a Simple JGo applet/application that demonstrates the use
 * of BasicNodes, both rectangular and elliptical, with the label in the middle.
 * <p>
 * The user can do all the standard mouse manipulations (select, move, copy, link)
 * and can insert new ones with the INSERT key, delete with the DELETE key,
 * cut/copy/paste (^X/^C/^V) if not an applet, and undo/redo (^Z/^Y).
 */
public class BasicApp implements JGoViewListener, JGoDocumentListener {
  public BasicApp(Shell shell) {
    myView = new JGoView(shell, SWT.V_SCROLL | SWT.H_SCROLL);
  }

  public static void main(String args[]) {
    try {
      Display display = new Display();
      Shell shell = new Shell(display);
      shell.setText("Basic JGo Application");
      FillLayout layout = new FillLayout();
      layout.type = SWT.VERTICAL;
      shell.setLayout(layout);
      
      new BasicApp(shell).init();

      shell.open();
      while (!shell.isDisposed()) {
        if (!display.readAndDispatch())
          display.sleep();
      }
    } catch (Throwable t) {
      System.err.println(t);
      t.printStackTrace();
      System.exit(1);
    }
  }

  public void init() {  // App initialization--only here do we do anything JGo specific
    // the view has a black background and magenta selection handles
    myView.setDragEnabled(true);
    myView.setDropEnabled(true);
    myView.setBackgroundColor(JGoBrush.ColorBlack);
    myView.setPrimarySelectionColor(JGoBrush.ColorMagenta);
    myView.setSecondarySelectionColor(myView.getPrimarySelectionColor());
    myView.setIncludingNegativeCoords(true);
    myView.setHidingDisabledScrollbars(true);
    myView.addKeyListener(new KeyAdapter() {
      public void keyPressed(KeyEvent evt) {
        handleKeyPressed(evt);
      }
    });
    // detect whenever a temporary link is added to the view so we can change its color
    myView.addViewListener(this);

    // initialize the document
    JGoDocument doc = myView.getDocument();
    // detect whenever a link is added to the document so we can modify its appearance
    doc.addDocumentListener(this);
    
    // create a couple of nodes and a link between them
    JGoBasicNode b1 = insertNode(new Point(50, 50), true);
    b1.setText("first");
    JGoBasicNode b2 = insertNode(new Point(130, 50), false);
    b2.setText("second");
    doc.addObjectAtTail(new JGoLink(b1.getPort(), b2.getPort()));
    
    // enable undo & redo
    doc.setUndoManager(new JGoUndoManager());
  }

  void handleKeyPressed(KeyEvent evt) {
    char ch = evt.character;
    int code = evt.keyCode;
    boolean ctrl = ((evt.stateMask & SWT.CTRL) != 0);
    boolean shift = ((evt.stateMask & SWT.SHIFT) != 0);
    if (code == SWT.DEL) {  // delete the currently selected objects
      myView.deleteSelection();
    } else if (code == SWT.INSERT) {   // insert a new node in the middle of the window
      Rectangle rectangle = myView.getViewRect();
      insertNode(new Point(rectangle.x + rectangle.width / 2, rectangle.y + rectangle.height / 2),
                 (myNodeCounter%2 == 0));
    } else if (code == SWT.HOME) {  // scroll so that the top-left-most object is at the top-left corner
      Rectangle docbounds = myView.getDocument().computeBounds();
      myView.setViewPosition(docbounds.x, docbounds.y);
    } else if (code == SWT.END) {  // scroll so that the bottom-right-most object is at the bottom-right corner
      Rectangle docbounds = myView.getDocument().computeBounds();
      Dimension viewsize = myView.getExtentSize();
      myView.setViewPosition(Math.max(docbounds.x, docbounds.x+docbounds.width-viewsize.width),
                             Math.max(docbounds.y, docbounds.y+docbounds.height-viewsize.height));
    } else if (ctrl && ch == 1) {  // Ctrl-A
      myView.selectAll();
    } else if (ctrl && ch == 3) {  // Ctrl-C
      myView.copy();
    } else if (ctrl && ch == 24) { // Ctrl-X
      myView.cut();
    } else if (ctrl && ch == 22) { // Ctrl-V
      myView.paste();
    } else if (ctrl && ch == 25) { // Ctrl-Y
      myView.getDocument().redo();
    } else if (ctrl && ch == 26) { // Ctrl-Z
      myView.getDocument().undo();
    }
  }

  // JGoViewListener:
  // Modify the temporary link created when drawing a new link
  // (the default black would not appear against the black background)
  public void viewChanged(JGoViewEvent evt) {
    if (evt.getHint() == JGoViewEvent.INSERTED) {
      if (evt.getJGoObject() instanceof JGoLink) {
        JGoLink link = (JGoLink)evt.getJGoObject();
        link.setPen(JGoPen.make(JGoPen.SOLID, 2, JGoBrush.ColorGray));
        link.setBrush(JGoBrush.gray);
        link.setArrowHeads(false, true);
      }
    } else if (evt.getHint() == JGoViewEvent.BACKGROUND_DOUBLE_CLICKED) {
      insertNode(evt.getPointDocCoords(), (myNodeCounter%2 == 0));
    }
  }

  // JGoDocumentListener:
  // If a link is added to the document, make sure it has a new color, is wider than
  // normal, and has an arrowhead at its "to" end
  public void documentChanged(JGoDocumentEvent evt) {
    if (evt.getHint() == JGoDocumentEvent.INSERTED && (evt.getJGoObject() instanceof JGoLink)) {
      JGoLink link = (JGoLink)evt.getJGoObject();
      link.setBrush(JGoBrush.makeStockBrush(getRandomColor(100)));  // brush fills the arrowhead
      link.setPen(JGoPen.make(JGoPen.SOLID, 2, link.getBrush().getColor()));  // pen controls the line and the outline of the arrowhead
      link.setArrowHeads(false, true);  // have an arrowhead at the "to" end
      link.setCubic(true);  // get a curved path for the link, controlled by Curviness property
      link.calculateStroke();  // replot the points of the stroke so that they are curved
    }
  }

  // Create a JGoBasicNode at a given point with a given shape.
  // Give it a randomly colored brush and pen, and a generated label string.
  // When changing a document, we do so within a transaction so that undo/redo works normally.
  JGoBasicNode insertNode(Point point, boolean rectangular) {
    JGoDocument doc = myView.getDocument();
    doc.startTransaction();
    JGoBasicNode bnode = new JGoBasicNode(Integer.toString(++myNodeCounter));
    bnode.setLabelSpot(JGoObject.Center);
    bnode.setLocation(point);
    if (rectangular)
      bnode.setDrawable(new JGoRectangle());
    bnode.setBrush(JGoBrush.makeStockBrush(getRandomColor(100)));
    bnode.setPen(JGoPen.make(JGoPen.SOLID, 2, getRandomColor(130)));
    bnode.getLabel().setEditable(true);
    doc.addObjectAtTail(bnode);
    doc.endTransaction("inserted node");
    return bnode;
  }

  RGB getRandomColor(int base) {
    return new RGB(base+(int)((250-base)*Math.random()),
                   base+(int)((250-base)*Math.random()),
                   base+(int)((250-base)*Math.random()));
  }

  private JGoView myView;
  private int myNodeCounter = 0;
}
