/*
 *  Copyright  Northwoods Software Corporation, 2000-2007. All Rights
 *  Reserved.
 *
 *  Restricted Rights: Use, duplication, or disclosure by the U.S.
 *  Government is subject to restrictions as set forth in subparagraph
 *  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
 *  52.227-14 Alt. III, as applicable.
 *
 */

package com.nwoods.jgo.examples;

import org.eclipse.swt.*;
import org.eclipse.swt.events.*;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.layout.*;
import org.eclipse.swt.widgets.*;
import org.eclipse.swt.custom.SashForm;
import org.eclipse.swt.dnd.*;
import com.nwoods.jgo.*;

/**
 * This is a Simple JGo applet/application that demonstrates the use
 * of JGoIconicNodes, adding objects from a palette without using Java drag-and-drop,
 * selecting the object under the mouse without clicking, and performing some
 * action with any single mouse click both on objects and in the background.
 * For nodes, the action brings up a dialog to modify the label.
 * For both nodes and links, a right-mouse-click deletes the object, after a
 * confirming dialog.
 * <p>
 * Users can just click at spots in the view in order to create new IconicNodes
 * at that location.  The node's icon is determined by the currently selected image
 * in the palette.  Unlike most examples, drag-and-drop is disabled from the palette.
 * When no image is selected in the palette, no JGoIconicNode is created when the user
 * clicks in the background.  The palette is also different than most because the
 * contents of the palette are just JGoImages, rather than whole nodes such as
 * JGoIconicNodes.
 * <p>
 * When the "Insert Mode" check box is turned off, clicking in the background
 * does not insert any new JGoIconicNodes.
 * <p>
 * As the mouse moves around the view, any object underneath the mouse is selected.
 * The status bar is updated with a description of the primary selection.
 * <p>
 * However, when the "Point to Select" mode is turned off, passing over an object
 * does not change the selection.  A single click selects an object, and a double-click
 * performs the action on the object.
 * <p>
 * Various keyboard commands are also defined--see the implementation of
 * handleKeyPressed for details.
 */
public class IconicApp implements JGoViewListener {
  public IconicApp(Shell shell) {
    shell.setLayout(new FormLayout());

    // First, the toolbar (Composite)
    Composite toolbar = new Composite(shell, SWT.NONE);
    toolbar.setLayout(new RowLayout(SWT.HORIZONTAL));
    Composite checkboxes = new Composite(toolbar, SWT.NONE);
    checkboxes.setLayout(new FillLayout(SWT.HORIZONTAL));
    // Add check boxes to the toolbar
    myInsertModeCheckBox = new Button(checkboxes, SWT.CHECK);
    myInsertModeCheckBox.setText("Insert Mode");
    myInsertModeCheckBox.setSelection(true);
    myPointToSelectModeCheckBox = new Button(checkboxes, SWT.CHECK);
    myPointToSelectModeCheckBox.setText("Point to Select");
    myPointToSelectModeCheckBox.setSelection(true);
    myLinkModeCheckBox = new Button(checkboxes, SWT.CHECK);
    myLinkModeCheckBox.setText("Linking Mode");
    myLinkModeCheckBox.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent selectionevent) {
        setLinkMode(myLinkModeCheckBox.getSelection());
      }
    });

    // Second, the SashForm holding the palette and the main view
    SashForm sash = new SashForm(shell, SWT.HORIZONTAL);
    myPalette = new JGoPalette(sash, SWT.LEFT | SWT.V_SCROLL);
    myView = new JGoView(sash, SWT.RIGHT | SWT.V_SCROLL | SWT.H_SCROLL) {
      public void doUncapturedMouseMove(int i, Point point, Point point1) {
        if (myPointToSelectModeCheckBox.getSelection())
          handleSelectionUnderMouse(point);
        super.doUncapturedMouseMove(i, point, point1);
      }
    };
    int[] weights = new int[2];
    weights[0] = 10;
    weights[1] = 90;
    sash.setWeights(weights);

    // Third, the status bar
    Composite statusbar = new Composite(shell, SWT.NONE);
    statusbar.setLayout(new FillLayout(SWT.HORIZONTAL));
    // add status label to status bar
    myStatusBarLabel = new Label(statusbar, SWT.NONE);
    myStatusBarLabel.setText(" ");

    FormData formData1 = new FormData();
    formData1.top = new FormAttachment(0, 3);
    formData1.left = new FormAttachment(0, 3);
    formData1.right = new FormAttachment(100, -3);
    toolbar.setLayoutData(formData1);

    FormData formData2 = new FormData();
    formData2.top = new FormAttachment(toolbar, 3);
    formData2.left = new FormAttachment(0, 0);
    formData2.right = new FormAttachment(100, 0);
    formData2.bottom = new FormAttachment(statusbar, -3);
    sash.setLayoutData(formData2);

    FormData formData3 = new FormData();
    formData3.left = new FormAttachment(0, 3);
    formData3.right = new FormAttachment(100, -3);
    formData3.bottom = new FormAttachment(100, -3);
    statusbar.setLayoutData(formData3);

    myView.addViewListener(this);
    myView.addKeyListener(new KeyAdapter() {
      public void keyPressed(KeyEvent evt) {
        handleKeyPressed(evt);
      }
    });
  }

  public static void main(String args[]) {
    try {
      Display display = new Display();
      Shell shell = new Shell(display);
      myShell = shell;
      shell.setText("Basic JGo Application");
      FormLayout layout = new FormLayout();
      shell.setLayout(layout);

      new IconicApp(shell).init();

      shell.open();
      while (!shell.isDisposed()) {
        if (!display.readAndDispatch())
          display.sleep();
      }
    } catch (Throwable t) {
      System.err.println(t);
      t.printStackTrace();
      System.exit(1);
    }
  }

  public void init() {
    // disable starting a drag-and-drop in the view
    // disable dragging from the palette and dropping in the view
    myView.setDragDropEnabled(false);  // ??
    myView.setIncludingNegativeCoords(true);
    myView.setHidingDisabledScrollbars(true);
    myView.getDocument().setUndoManager(new JGoUndoManager());
    myPalette.setInternalMouseActions(DND.DROP_NONE);
    myPalette.setDragDropEnabled(false);

    // in this example, the palette just holds images, not whole nodes
    JGoImage img = new JGoImage();
    img.setResizable(false);
    img.loadImage((IconicApp.class).getResource("demo1SWT/star.gif"), true);
    img.setSize(32, 32);
    myPalette.getDocument().addObjectAtTail(img);
    img = new JGoImage();
    img.setResizable(false);
    img.loadImage((IconicApp.class).getResource("demo1SWT/doc.gif"), true);
    img.setSize(32, 32);
    myPalette.getDocument().addObjectAtTail(img);
    myPalette.selectObject(img);
  }

  // As the mouse moves over objects, we select them, thereby deselecting any
  // other object.  As the mouse moves over the background, we clear the selection,
  // unless there's already a multiple selection specified by the user.
  void handleSelectionUnderMouse(Point point) {
    int num = myView.getSelection().getNumObjects();
    JGoObject obj = myView.pickDocObject(point, true);
    if (obj != null && num <= 1)
      myView.selectObject(obj);
    else if (obj == null && num == 1)
      myView.getSelection().clearSelection();
  }

  void handleKeyPressed(KeyEvent evt) {
    char ch = evt.character;
    int code = evt.keyCode;
    boolean ctrl = ((evt.stateMask & SWT.CTRL) != 0);
    boolean shift = ((evt.stateMask & SWT.SHIFT) != 0);
    if (code == SWT.DEL) {  // delete the currently selected objects
      myView.deleteSelection();
    } else if (code == SWT.INSERT) {   // insert a new node in the middle of the window
      Rectangle rectangle = myView.getViewRect();
      insertNode(new Point(rectangle.x + rectangle.width / 2, rectangle.y + rectangle.height / 2));
    } else if (code == SWT.HOME) {  // scroll so that the top-left-most object is at the top-left corner
      Rectangle docbounds = myView.getDocument().computeBounds();
      myView.setViewPosition(docbounds.x, docbounds.y);
    } else if (code == SWT.END) {  // scroll so that the bottom-right-most object is at the bottom-right corner
      Rectangle docbounds = myView.getDocument().computeBounds();
      Dimension viewsize = myView.getExtentSize();
      myView.setViewPosition(Math.max(docbounds.x, docbounds.x+docbounds.width-viewsize.width),
                             Math.max(docbounds.y, docbounds.y+docbounds.height-viewsize.height));
    } else if (ctrl && ch == 1) {  // Ctrl-A
      myView.selectAll();
    } else if (ctrl && ch == 3) {  // Ctrl-C
      myView.copy();
    } else if (ctrl && ch == 24) { // Ctrl-X
      myView.cut();
    } else if (ctrl && ch == 22) { // Ctrl-V
      myView.paste();
    } else if (ctrl && ch == 25) { // Ctrl-Y
      myView.getDocument().redo();
    } else if (ctrl && ch == 26) { // Ctrl-Z
      myView.getDocument().undo();
    }
  }

  // When a link is drawn by the user, make it wide and blue with an arrowhead.
  // A single click on a node starts editing the node--you can substitute other operations
  // if you desire.  A single right mouse click causes a different action, in this demo,
  // a prompted deletion.
  // When PointToSelect mode is false, these actions occur on a double-click instead.
  // A click in the background inserts a new node at that point.
  // We also handle selection gained and selection lost view events to maintain
  // the status bar's description of the current object, the primary selection.
  public void viewChanged(JGoViewEvent evt) {
    if (evt.getHint() == JGoViewEvent.LINK_CREATED && (evt.getJGoObject() instanceof JGoLink)) {
      JGoLink link = (JGoLink)evt.getJGoObject();
      link.setPen(JGoPen.make(JGoPen.SOLID, 2, JGoBrush.ColorBlue));
      link.setArrowHeads(false, true);
    } else if (evt.getHint() == JGoViewEvent.CLICKED) {
      if (myPointToSelectModeCheckBox.getSelection()) {
        if ((evt.getModifiers() & SWT.BUTTON3) != 0)
          deleteObject(evt.getJGoObject());
        else
          editObject(evt.getJGoObject());
      }
    } else if (evt.getHint() == JGoViewEvent.DOUBLE_CLICKED) {
      if (!myPointToSelectModeCheckBox.getSelection()) {
        if ((evt.getModifiers() & SWT.BUTTON3) != 0)
          deleteObject(evt.getJGoObject());
        else
          editObject(evt.getJGoObject());
      }
    } else if (evt.getHint() == JGoViewEvent.BACKGROUND_CLICKED) {
      if (myInsertModeCheckBox.getSelection() &&
          myPalette.getSelection().getPrimarySelection() != null &&
          myView.getSelection().getNumObjects() == 0 &&
          (evt.getModifiers() & SWT.BUTTON3) == 0) {  // not right-mouse-button
        insertNode(evt.getPointDocCoords());
      }
    } else if (evt.getHint() == JGoViewEvent.SELECTION_GAINED ||
               evt.getHint() == JGoViewEvent.SELECTION_LOST) {
      describeCurrentObject();
    }
  }

  // Create an JGoIconicNode at a given point with an icon whose image is selected in the palette.
  // The node's Port is Visible when linking mode is true so that users can draw new links.
  // When changing a document, we do so within a transaction so that undo/redo works normally.
  void insertNode(Point point) {
    JGoDocument doc = myView.getDocument();
    doc.startTransaction();
    JGoObject obj = myPalette.getSelection().getPrimarySelection();
    JGoIconicNode node = new JGoIconicNode(Integer.toString(++myNodeCounter));
    // copy the JGoImage
    JGoImage img = (JGoImage)obj.copyObject(new JGoCopyMap());
    // initialize the new node's Icon and its other properties
    node.setIcon(img);
    node.setLocation(point);
    node.getLabel().setEditable(false);
    node.getPort().setBoundingRect(node.getIcon().getBoundingRect());
    node.getPort().setVisible(myLinkModeCheckBox.getSelection());
    doc.addObjectAtTail(node);
    doc.endTransaction("inserted node");
  }

  // For this example, editing a node just lets the user change the label.
  // You can substitute much more descriptive dialogs for the information appropriate
  // to your application, or perform other actions.
  void editObject(JGoObject obj) {
    if (obj == null) return;
    obj = obj.getParentNode();
    if (obj instanceof JGoIconicNode) {
      JGoIconicNode node = (JGoIconicNode)obj;
      String str = node.getLabel().getText();
      if (str != null && !str.equals(node.getLabel().getText())) {
        JGoDocument doc = myView.getDocument();
        doc.startTransaction();
        node.getLabel().setText(str);
        doc.endTransaction("changed label");
      }
    }  // but don't do anything for a link
  }

  // An alternative action, for right-mouse-button events.
  // You might substitute bringing up a popup menu with different commands.
  void deleteObject(JGoObject obj) {
    if (obj == null) return;
    MessageBox dlg = new MessageBox(myShell, SWT.ICON_QUESTION | SWT.YES | SWT.NO);
    dlg.setMessage("Delete selection?");
    if (dlg.open() == SWT.YES) {
      myView.deleteSelection();
    }
  }

  // Look at the primary selection and update the status bar correspondingly.
  void describeCurrentObject() {
    JGoObject obj = myView.getSelection().getPrimarySelection();
    if (obj == null) {
      myStatusBarLabel.setText(" ");
    } else {
      if (obj instanceof JGoIconicNode) {
        JGoIconicNode node = (JGoIconicNode)obj;
        myStatusBarLabel.setText("node " + node.getLabel().getText());
      } else if (obj instanceof JGoLink) {
        JGoLink link = (JGoLink)obj;
        JGoIconicNode from = (JGoIconicNode)link.getFromPort().getParent();
        JGoIconicNode to = (JGoIconicNode)link.getToPort().getParent();
        String msg = "link from " + from.getLabel().getText() + " to " + to.getLabel().getText();
        myStatusBarLabel.setText(msg);
      }
    }
  }

  // Modify all of the IconicNodes so that their ports are Visible if users need to be able to
  // draw links between them; otherwise the ports are not Visible, so that a mouse down on the
  // icon does not start drawing a new link.
  // Note that we cause all these document changes to avoid being recorded by the UndoManager.
  void setLinkMode(boolean flag) {
    JGoDocument doc = myView.getDocument();
    doc.setSkipsUndoManager(true);
    JGoListPosition pos = doc.getFirstObjectPos();
    while (pos != null) {
      JGoObject obj = doc.getObjectAtPos(pos);
      pos = doc.getNextObjectPosAtTop(pos);
      if (obj instanceof JGoIconicNode) {
        JGoIconicNode node = (JGoIconicNode)obj;
        node.getPort().setVisible(flag);
      }
    }
    doc.setSkipsUndoManager(false);
  }

  private JGoView myView;
  private JGoPalette myPalette;
  private static Shell myShell;
  private Button myPointToSelectModeCheckBox;
  private Button myInsertModeCheckBox;
  private Button myLinkModeCheckBox;
  private Label myStatusBarLabel;
  private int myNodeCounter = 0;
}
