/*
 *  Copyright  Northwoods Software Corporation, 1998-2007. All Rights
 *  Reserved.
 *
 *  Restricted Rights: Use, duplication, or disclosure by the U.S.
 *  Government is subject to restrictions as set forth in subparagraph
 *  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
 *  52.227-14 Alt. III, as applicable.
 *
 */

package com.nwoods.jgo.examples.demo1SWT;

import org.eclipse.swt.*;
import org.eclipse.swt.dnd.*;
import org.eclipse.swt.events.*;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.layout.*;
import org.eclipse.swt.printing.*;
import org.eclipse.swt.widgets.*;
import java.util.*;
import com.nwoods.jgo.*;
import com.nwoods.jgo.examples.*;

public class Demo1View extends JGoView
{
  Demo1View(Composite parent, int style)
  {
    super(parent, style | SWT.BORDER);
    setIncludingNegativeCoords(true);
    setHidingDisabledScrollbars(true);

    Transfer[] viewTypes = new Transfer[] { JGoTransfer.getInstance(), TextTransfer.getInstance() };
    getDropTarget().setTransfer(viewTypes);
  }

  // new kinds of "modes"
  public static final int MouseStateDrawingStroke = MouseStateLast+1;

  public void doCancelMouse() {
    myMouseUpDocPoint.x = -9999;
    myMouseUpDocPoint.y = -9999;
    if (getState() == MouseStateDrawingStroke) {
      cancelDrawingStroke();
    } else {
      super.doCancelMouse();
    }
  }

  public void onKeyEvent(KeyEvent evt) {
    int t = evt.keyCode;
    if (t == SWT.ESC) {
      doCancelMouse();
    } else if (t == SWT.CR && getState() == MouseStateDrawingStroke) {
      JGoStroke myStroke = (JGoStroke)getCurrentObject();
      if (myStroke != null && myStroke.getNumPoints() > 2) {
        // stroke always includes a point where the mouse is, which isn't
        // supposed to be included in the finished JGoStroke
        finishDrawingStroke();
      } else {
        doCancelMouse();
      }
    } else {
      super.onKeyEvent(evt);
    }
  }

  // This implements additional mouse behaviors on mouse down:
  // - start or continue creating a stroke with a user-defined path
  // - if on a record node field that is not Draggable, then we
  //   start drawing a link from that field's port
  public boolean doMouseDown(int modifiers, Point dc, Point vc)
  {
    if (getState() == MouseStateDrawingStroke) {
      addPointToDrawingStroke(modifiers, dc, vc);
      return true;
    }
    JGoObject obj = pickDocObject(dc, true);
    if (obj != null) {
      JGoObject field = obj;
      if (!field.isDraggable() &&
        field.getParent() != null &&
        field.getParent() instanceof ListArea &&
        field.getParent().getParent() != null &&
        field.getParent().getParent() instanceof RecordNode) {
        RecordNode node = (RecordNode)field.getParent().getParent();
        int idx = node.findItem(field);
        if (idx >= 0) {
          JGoPort port = node.getLeftPort(idx);
          if (port == null)
            port = node.getRightPort(idx);
          if (port != null &&
            startNewLink(port, dc)) {
            return true;
          }
        }
      }
    }
    // otherwise implement the default behavior
    return super.doMouseDown(modifiers, dc, vc);
  }

  // This implements additional mouse behaviors on mouse down:
  // - when creating a link with a user-defined path, have the end follow
  //   the mouse pointer between clicks
  public boolean doMouseMove(int modifiers, Point dc, Point vc)
  {
    if (getState() == MouseStateDrawingStroke) {
      followPointerForDrawingStroke(modifiers, dc, vc);
      return true;
    }
    return super.doMouseMove(modifiers, dc, vc);
  }

  private Point myMouseUpDocPoint = new Point(0, 0);

  // This implements additional mouse behaviors on mouse up:
  // - mouse up must be ignored when the user is specifying stroke points
  // - if it's a right button mouse down bring up a properties
  //   dialog for the object
  public boolean doMouseUp(int modifiers, Point dc, Point vc)
  {
    myMouseUpDocPoint.x = dc.x;
    myMouseUpDocPoint.y = dc.y;

    if (getState() == MouseStateDrawingStroke) {
      // don't need to do anything
      return false;
    }
    if ((modifiers & SWT.BUTTON3) != 0) {
      JGoObject obj = pickDocObject(dc, true);
      if (obj != null) {  // context click on an object
        selectObject(obj);
        if (obj instanceof JGoLink) {
          JGoLink link = (JGoLink)obj;
          Menu popup = new Menu(this);
          MenuItem insertPointMI = new MenuItem(popup, SWT.NONE);
          insertPointMI.setText("Insert Point");
          insertPointMI.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) { insertPointIntoLink(); }
          });
          if (link.getNumPoints() > (link.isOrthogonal() ? 6 : 2)) {
            MenuItem removeSegmentMI = new MenuItem(popup, SWT.NONE);
            removeSegmentMI.setText("Remove Segment");
            removeSegmentMI.addSelectionListener(new SelectionAdapter() {
              public void widgetSelected(SelectionEvent e) { removeSegmentFromLink(); }
            });
          }
          // coords need to be relative to Display
          popup.setLocation(toDisplay(vc));
          popup.setVisible(true);
        }
        return true;
      }
    }
    // otherwise implement the default behavior
    return super.doMouseUp(modifiers, dc, vc);
  }

  // don't start drawing a new link if the port is selectable
  public boolean startNewLink(JGoPort port, Point dc) {
    if (port.isSelectable()) return false;
    return super.startNewLink(port, dc);
  }

  public void startDrawingStroke() {
    setState(MouseStateDrawingStroke);
    getDocument().startTransaction();
    setCurrentObject(null);
    setCursorType(SWT.CURSOR_CROSS);
  }

  void cancelDrawingStroke() {
    if (!isFocusControl())
      return;
    if (getCurrentObject() != null) {
      removeObject(getCurrentObject());
      setCurrentObject(null);
    }
    setCursorType(getDefaultCursorType());
    getDocument().endTransaction(false);
    setState(MouseStateNone);
  }

  void addPointToDrawingStroke(int modifiers, Point dc, Point vc) {
    // keep adding points to the stroke representing the future link
    JGoStroke s = (JGoStroke)getCurrentObject();
    if (s == null) {
      // create a new stroke starting at or near the first mouse down point
      s = new JGoStroke();
      s.addPoint(dc);
      addObjectAtTail(s);
      setCurrentObject(s);
    }
    s.addPoint(dc);
  }

  void followPointerForDrawingStroke(int modifiers, Point dc, Point vc) {
    JGoStroke myStroke = (JGoStroke)getCurrentObject();
    if (myStroke != null) {
      int numpts = myStroke.getNumPoints();
      myStroke.setPoint(numpts-1, dc);
    }
  }

  void finishDrawingStroke() {
    JGoStroke myStroke = (JGoStroke)getCurrentObject();
    if (myStroke != null) {
      // get rid of the stroke from the view
      removeObject(myStroke);
      // remove last point, which had been following the pointer
      myStroke.removePoint(myStroke.getNumPoints()-1);
      getDocument().getDefaultLayer().addObjectAtTail(myStroke);
      selectObject(myStroke);
      setCurrentObject(null);
      setCursorType(getDefaultCursorType());
      getDocument().endTransaction("created stroke");
      setState(MouseStateNone);
    }
  }


  // popup menu commands

  void insertPointIntoLink()
  {
    if (getSelection().getPrimarySelection() instanceof JGoLink) {
      JGoLink s = (JGoLink)getSelection().getPrimarySelection();
      int i = s.getSegmentNearPoint(myMouseUpDocPoint);
      if (s.getNumPoints() > 3) {
        if (i < 1)
          i = 1;  // don't add to first segment
        else if (i >= s.getNumPoints()-2)
          i = s.getNumPoints()-3;  // don't add to last segment
      }
      Point a = s.getPoint(i);
      Point b = s.getPoint(i+1);
      Point closest = new Point((a.x+b.x)/2, (a.y+b.y)/2);
      getDocument().startTransaction();
      s.insertPoint(i+1, closest);
      if (s.isOrthogonal())  // when orthogonal, gotta insert two points
        s.insertPoint(i+1, closest);
      getSelection().toggleSelection(s);
      selectObject(s);
      getDocument().endTransaction("inserted point into link stroke");
    }
  }

  void removeSegmentFromLink()
  {
    if (getSelection().getPrimarySelection() instanceof JGoLink) {
      JGoLink s = (JGoLink)getSelection().getPrimarySelection();
      int i = s.getSegmentNearPoint(myMouseUpDocPoint);
      getDocument().startTransaction();
        if (s.isOrthogonal()) {  // will have at least 7 points
          // don't remove either first two or last two segments
          i = Math.max(i, 2);
          i = Math.min(i, s.getNumPoints()-5);
          Point a = s.getPoint(i);
          Point b = s.getPoint(i+1);
          s.removePoint(i);
          // to maintain orthogonality, gotta remove two points
          s.removePoint(i);
          // now fix up following point to maintain orthogonality
//          Point c = new Point(s.getPoint(i));
          Point c = new Point(s.getPoint(i).x, s.getPoint(i).y);
          if (a.x == b.x) {
            c.y = a.y;
          } else {
            c.x = a.x;
          }
          s.setPoint(i, c);
        } else {  // will have at least 3 points
          i = Math.max(i, 1);  // don't remove point 0
          i = Math.min(i, s.getNumPoints()-2);  // don't remove last point
          s.removePoint(i);
        }
      getSelection().toggleSelection(s);
      selectObject(s);
      getDocument().endTransaction("removed segment from link stroke");
    }
  }

  // demonstrate how to detect a double-click in the background--i.e., not on any object
  // (for single-click in background, see doBackgroundClick below...)
  public boolean doMouseDblClick(int modifiers, Point dc, Point vc)
  {
    boolean result = super.doMouseDblClick(modifiers, dc, vc);
    if (!result && pickDocObject(dc,false) == null) {
      getDocument().startTransaction();
      LimitedNode n = new LimitedNode("limited");
      n.setLocation(dc);
      getDocument().addObjectAtTail(n);
      getDocument().endTransaction("added LimitedNode in background");
    }
    return result;
  }

  // If a right button mouse click, maybe do something
  public void doBackgroundClick(int modifiers, Point dc, Point vc)
  {
    // System.err.println("single background click " + Integer.toString(modifiers));
    if ((modifiers & SWT.BUTTON3) != 0) {
      Menu popup = new Menu(this);
      MenuItem miProps = new MenuItem(popup, SWT.NONE);
      miProps.setText("Properties");
      miProps.addSelectionListener(new SelectionAdapter() {
        public void widgetSelected(SelectionEvent e) { inspect(); }
      });
      if (canPaste()) {
        MenuItem miPaste = new MenuItem(popup, SWT.NONE);
        miPaste.setText("Paste");
        final Point pastePoint = new Point(dc.x, dc.y);
        miPaste.addSelectionListener(new SelectionAdapter() {
          public void widgetSelected(SelectionEvent e) { pasteAtPoint(pastePoint); }
        });
      }
      popup.setLocation(toDisplay(vc));
      popup.setVisible(true);
    } else {
      super.doBackgroundClick(modifiers, dc, vc);
    }
  }

  // inspect this view
  public void inspect() {
    Inspector.inspect(getDisplay(), getShell(), this);
    Composite inspector = Inspector.getInspector();
    if (inspector != null && !inspector.isVisible())
      inspector.setVisible(true);
  }

  // call paste() to copy objects from the clipboard, and then move
  // the selected new objects to the popup-menu point
  public void pasteAtPoint(Point dc) {
    getDocument().startTransaction();
    paste();
    if (!getSelection().isEmpty()) {
      Rectangle bounds = JGoDocument.computeBounds(getSelection());
      moveSelection(getSelection(), 0, dc.x-bounds.x, dc.y-bounds.y, EventMouseUp);
    }
    getDocument().endTransaction("popup paste");
  }

  // don't allow text to be dropped onto a JGoObject,
  // but allow it to be dropped elsewhere in the view
  public int computeAcceptableDrop(DropTargetEvent e) {
    if (TextTransfer.getInstance().isSupportedType(e.currentDataType)) {
      Point p = toControl(new Point(e.x, e.y));
      convertViewToDoc(p);
      JGoObject obj = pickDocObject(p, true);
      if (obj != null)
        return DND.DROP_NONE;
      else
        return DND.DROP_COPY;
    } else {
      return super.computeAcceptableDrop(e);
    }
  }

  // handle drops of text, as well as of the usual JGoSelection
  public void onExternalDrop(DropTargetEvent e) {
    JGoDocument doc = getDocument();
    doc.startTransaction();
    if (TextTransfer.getInstance().isSupportedType(e.currentDataType)) {  // handle text
      String s = (String)e.data;
      JGoText t = new JGoText();
      t.setText(s);
      t.setAlignment(JGoObject.Center);
      Point p = toControl(new Point(e.x, e.y));
      convertViewToDoc(p);
      t.setLocation(p);
      doc.addObjectAtTail(t);
      getSelection().selectObject(t);
      fireUpdate(JGoViewEvent.EXTERNAL_OBJECTS_DROPPED, 0, null);
      doc.endTransaction("dropped text");
    } else if (doDrop(e, null)) {  // handle drop of JGoSelection (i.e. JGoTransfer)
      fireUpdate(JGoViewEvent.EXTERNAL_OBJECTS_DROPPED, 0, null);
      doc.endTransaction(getEditPresentationName(6));
    } else {
      e.detail = DND.DROP_NONE;
      doc.endTransaction(false);
    }
  }

  // this is a helper function for deciding whether
  // to create a labeled link or not
  public boolean makeLabeledLinkForPort(JGoPort p)
  {
    if (p.getParent() instanceof JGoBasicNode)
      return true;
    JGoArea area = p.getParent();
    if (area != null &&
        (area.getFlags() & Demo1.flagMultiSpotNode) != 0)
      return true;
    return false;
  }

  // This method is called by JGoView when the user has
  // performed the gestures for creating a new link between
  // two ports.  We need to create the link, add it to the
  // document, and finish the transaction.
  public void newLink(JGoPort from, JGoPort to)
  {
    JGoLink l = null;
    if (makeLabeledLinkForPort(from) &&
        makeLabeledLinkForPort(to)) {
      JGoLabeledLink link = new JGoLabeledLink(from, to);
      l = link;

      JGoLinkLabel text;
      if (!(from.getParent() instanceof JGoBasicNode)) {
        text = new JGoLinkLabel("from");
        text.setAlignment(JGoText.ALIGN_LEFT);
        text.setSelectable(true);
        text.setEditOnSingleClick(true);
        link.setFromLabel(text);
      }

      text = new JGoLinkLabel("middle");
      text.setAlignment(JGoText.ALIGN_CENTER);
      text.setSelectable(true);
      text.setEditable(true);
      text.setEditOnSingleClick(true);
      link.setMidLabel(text);

      if (!(to.getParent() instanceof JGoBasicNode)) {
        text = new JGoLinkLabel("to");
        text.setAlignment(JGoText.ALIGN_RIGHT);
        text.setEditOnSingleClick(true);
        link.setToLabel(text);
      }
    } else {
      l = new JGoLink(from, to);
    }

    JGoSubGraphBase.reparentToCommonSubGraph(l, from, to, true, from.getDocument().getLinksLayer());

    fireUpdate(JGoViewEvent.LINK_CREATED, 0, l);
    from.getDocument().endTransaction("new link");
  }

  public double getPrintScale(Graphics2D g2, Printer p) {
    // three different options:

    // (A) always print at standard scale
    /*
     return 1.0d;
     */

    // (B) print using view's currently selected scale
    return getScale();

    // (C) scale to fit printed page
    /*
     Rectangle pageRect = getPrintPageRect(g2, pf);
     Dimension docSize = getPrintDocumentSize();
     // make sure it doesn't get scaled too much! (especially if no objects in document)
     docSize.width = Math.max(docSize.width, 50);
     docSize.height = Math.max(docSize.height, 50);
     double hratio = pageRect.width / (double)docSize.width;
     double vratio = pageRect.height / (double)docSize.height;
     return Math.min(hratio, vratio);
     */
  }
}
